"""
2.2: Centered Grid.
^^^^^^^^^^^^^^^^^^^

Geophysics Preprocessing builds on the centered grid
(https://github.com/cgre-aachen/gempy/blob/master/notebooks/tutorials/ch1-3-Grids.ipynb)
to precompute the constant part of forward physical computations as for
example gravity:

.. math::

    F_z = G_{\\rho} ||| x \\ln(y+r) + y \\ln (x+r) - z \\arctan (\\frac{x y}{z r}) |^{x_2}_{x_1}|^{y_2}_{y_1}|^{
    z_2}_{z_1}


where we can compress the grid dependent terms as

.. math::

    t_z = ||| x \ln (y+r) + y \ln (x+r)-z \\arctan ( \\frac{x y}{z r} ) |^{x_2}_{x_1}|^{y_2}_{y_1}|^{z_2}_{z_1}

By doing this decomposition an keeping the grid constant we can compute
the forward gravity by simply operate:

.. math::

    F_z = G_{\\rho} \cdot t_z

"""


# %%

# Importing gempy
from gempy_legacy.assets.geophysics import GravityPreprocessing

# Aux imports
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

np.random.seed(1515)
pd.set_option('display.precision', 2)

# %% 
g = GravityPreprocessing()

# %% 
kernel_centers, kernel_dxyz_left, kernel_dxyz_right = g.create_irregular_grid_kernel(resolution=[10, 10, 20],
                                                                                     radius=100)

# %%
# ``create_irregular_grid_kernel`` will create a constant kernel around
# the point 0,0,0. This kernel will be what we use for each device.
# 

# %% 
kernel_centers

# %%
# :math:`t_z` is only dependent on distance and therefore we can use the
# kerenel created on the previous cell
# 

# %% 
tz = g.set_tz_kernel(resolution=[10, 10, 20], radius=100)
tz

# %%
# To compute tz we also need the edges of each voxel. The distance to the
# edges are stored on ``kernel_dxyz_left`` and ``kernel_dxyz_right``. We
# can plot all the data as follows:
# 

# %% 
a, b, c = kernel_centers, kernel_dxyz_left, kernel_dxyz_right

# %% 
fig = plt.figure(figsize=(13, 7))
plt.quiver(a[:, 0].reshape(11, 11, 21)[5, :, :].ravel(),
           a[:, 2].reshape(11, 11, 21)[:, 5, :].ravel(),
           np.zeros(231),
           tz.reshape(11, 11, 21)[5, :, :].ravel(), label='$t_z$', alpha=.3
           )

plt.plot(a[:, 0].reshape(11, 11, 21)[5, :, :].ravel(),
         a[:, 2].reshape(11, 11, 21)[:, 5, :].ravel(), 'o', alpha=.3, label='Centers')

plt.plot(a[:, 0].reshape(11, 11, 21)[5, :, :].ravel() - b[:, 0].reshape(11, 11, 21)[5, :, :].ravel(),
         a[:, 2].reshape(11, 11, 21)[:, 5, :].ravel(), '.', alpha=.3, label='Lefts')

plt.plot(a[:, 0].reshape(11, 11, 21)[5, :, :].ravel(),
         a[:, 2].reshape(11, 11, 21)[:, 5, :].ravel() - b[:, 2].reshape(11, 11, 21)[:, 5, :].ravel(), '.', alpha=.6,
         label='Ups')

plt.plot(a[:, 0].reshape(11, 11, 21)[5, :, :].ravel() + c[:, 0].reshape(11, 11, 21)[5, :, :].ravel(),
         a[:, 2].reshape(11, 11, 21)[:, 5, :].ravel(), '.', alpha=.3, label='Rights')

plt.plot(a[:, 0].reshape(11, 11, 21)[5, :, :].ravel(),
         a[:, 2].reshape(11, 11, 21)[:, 5, :].ravel() + c[:, 2].reshape(11, 11, 21)[5, :, :].ravel(), '.', alpha=.3,
         label='Downs')

plt.xlim(-200, 200)
plt.ylim(-200, 0)
plt.legend()
plt.show()

# %%
# Just the quiver:
# 

# %%
fig = plt.figure(figsize=(13, 7))
plt.quiver(a[:, 0].reshape(11, 11, 21)[5, :, :].ravel(),
           a[:, 2].reshape(11, 11, 21)[:, 5, :].ravel(),
           np.zeros(231),
           tz.reshape(11, 11, 21)[5, :, :].ravel()
           )
plt.show()

# %%
# Remember this is happening always in 3D:
# 

# %%
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')

ax.scatter(a[:, 0], a[:, 1], a[:, 2], c=tz)

ax.set_xlabel('X Label')
ax.set_ylabel('Y Label')
ax.set_zlabel('Z Label')
plt.show()


"""
Perth basin.
============
"""
import os

# Importing GemPy
import gempy_legacy as gp

# Importing auxiliary libraries
import matplotlib
matplotlib.rcParams['figure.figsize'] = (20.0, 10.0)
os.environ["aesara_FLAGS"] = "mode=FAST_RUN,device=cuda"

# %%
data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
geo_model = gp.create_model('Perth_Basin')

# %% 
gp.init_data(geo_model,
             extent=[337000, 400000, 6640000, 6710000, -18000, 1000],
             resolution=[100, 100, 100],
             path_i=data_path + "/data/input_data/Perth_basin/Paper_GU2F_sc_faults_topo_Points.csv",
             path_o=data_path + "/data/input_data/Perth_basin/Paper_GU2F_sc_faults_topo_Foliations.csv")

# %%
geo_model.surfaces

# %% 
del_surfaces = ['Cadda', 'Woodada_Kockatea', 'Cattamarra']

# %% 
geo_model.delete_surfaces(del_surfaces, remove_data=True)

# %% 
# %debug

# %% 
geo_model.stack

# %% 
gp.map_stack_to_surfaces(geo_model,
                          {"fault_Abrolhos_Transfer": ["Abrolhos_Transfer"],
                           "fault_Coomallo": ["Coomallo"],
                           "fault_Eneabba_South": ["Eneabba_South"],
                           "fault_Hypo_fault_W": ["Hypo_fault_W"],
                           "fault_Hypo_fault_E": ["Hypo_fault_E"],
                           "fault_Urella_North": ["Urella_North"],
                           "fault_Urella_South": ["Urella_South"],
                           "fault_Darling": ["Darling"],
                           "Sedimentary_Series": ['Cretaceous',
                                                  'Yarragadee',
                                                  'Eneabba',
                                                  'Lesueur',
                                                  'Permian']
                           })

# %%
geo_model.series

# %% 
order_series = ["fault_Abrolhos_Transfer",
                "fault_Coomallo",
                "fault_Eneabba_South",
                "fault_Hypo_fault_W",
                "fault_Hypo_fault_E",
                "fault_Urella_North",
                "fault_Darling",
                "fault_Urella_South",
                "Sedimentary_Series", 'Basement']

geo_model.reorder_series(order_series)

# %%
# Drop input data from the deleted series:
# 

# %% 
geo_model.surface_points.df.dropna(inplace=True)
geo_model.orientations.df.dropna(inplace=True)

# %%
# Select which series are faults
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 

# %% 
geo_model.faults

# %% 
geo_model.set_is_fault(["fault_Abrolhos_Transfer",
                        "fault_Coomallo",
                        "fault_Eneabba_South",
                        "fault_Hypo_fault_W",
                        "fault_Hypo_fault_E",
                        "fault_Urella_North",
                        "fault_Darling",
                        "fault_Urella_South"])

# %%
# Fault Network
# -------------
# 

# %% 
geo_model.faults.faults_relations_df

# %% 
fr = geo_model.faults.faults_relations_df.values

# %% 
fr[:, :-2] = False
fr

# %% 
geo_model.set_fault_relation(fr)

# %% 
# %matplotlib inline
gp.plot_2d(geo_model, direction=['z'])

# %% 
geo_model.set_topography(source='random')

# %% 
gp.plot_3d(geo_model)

# %% 
interp_data = gp.set_interpolator(geo_model,
                                  compile_aesara=True,
                                  aesara_optimizer='fast_run', gradient=False,
                                  dtype='float32')

# %% 
gp.compute_model(geo_model)

# %% 
gp.plot_2d(geo_model, cell_number=[25])

# %% 
gp.plot_2d(geo_model, cell_number=[25], series_n=-1, show_scalar=True)

# %% 
gp.plot_2d(geo_model, cell_number=[12], direction=["y"], show_data=True, show_topography=True)

# %%
# sphinx_gallery_thumbnail_number = 6
gp.plot_3d(geo_model, show_topography=True)

# %%
# Times
# -----
# 
# Fast run
# ^^^^^^^^
# 
# -  1M voxels:
# 
#    -  CPU: intel® Core™ i7-7700HQ CPU @ 2.80GHz × 8 15 s ± 1.02 s per
#       loop (mean ± std. dev. of 7 runs, 1 loop each)
#    -  GPU (4gb) not enough memmory
#    -  Ceres 1M voxels 2080 851 ms
# 
# -  250k voxels
# 
#    -  GPU 1050Ti: 3.11 s ± 11.8 ms per loop (mean ± std. dev. of 7 runs,
#       1 loop each)
#    -  CPU: intel® Core™ i7-7700HQ CPU @ 2.80GHz × 8 2.27 s ± 47.3 ms
#    -  
# 
# Fast Compile
# ^^^^^^^^^^^^
# 
# -  250k voxels
# 
#    -  GPU 1050Ti: 3.7 s ± 11.8 ms per loop (mean ± std. dev. of 7 runs,
#       1 loop each)
#    -  CPU: intel® Core™ i7-7700HQ CPU @ 2.80GHz × 8 14.2 s ± 51.1 ms per
#       loop (mean ± std. dev. of 7 runs, 1 loop each)
# 

# %% 
# %%timeit
# gp.compute_model(geo_model)

# %% 
# ver = np.load('ver.npy')
# sim = np.load('sim.npy')
# lith_block = np.load('lith.npy')

gp.save_model(geo_model)



try:
    from pandas.plotting._tools import *
except:
    #TODO this is a quick fix for #38
    from pandas.plotting._matplotlib.tools import *


def _subplots(
    naxes=None,
    sharex=False,
    sharey=False,
    squeeze=True,
    subplot_kw=None,
    ax=None,
    layout=None,
    layout_type="box",
    **fig_kw
):
    """Create a figure with a set of subplots already made.

    This utility wrapper makes it convenient to create common layouts of
    subplots, including the enclosing figure object, in a single call.

    Keyword arguments:

    naxes : int
      Number of required axes. Exceeded axes are set invisible. Default is
      nrows * ncols.

    sharex : bool
      If True, the X axis will be shared amongst all subplots.

    sharey : bool
      If True, the Y axis will be shared amongst all subplots.

    squeeze : bool

      If True, extra dimensions are squeezed out from the returned axis object:
        - if only one subplot is constructed (nrows=ncols=1), the resulting
        single Axis object is returned as a scalar.
        - for Nx1 or 1xN subplots, the returned object is a 1-d numpy object
        array of Axis objects are returned as numpy 1-d arrays.
        - for NxM subplots with N>1 and M>1 are returned as a 2d array.

      If False, no squeezing is done: the returned axis object is always
      a 2-d array containing Axis instances, even if it ends up being 1x1.

    subplot_kw : dict
      Dict with keywords passed to the add_subplot() call used to create each
      subplots.

    ax : Matplotlib axis object, optional

    layout : tuple
      Number of rows and columns of the subplot grid.
      If not specified, calculated from naxes and layout_type

    layout_type : {'box', 'horizontal', 'vertical'}, default 'box'
      Specify how to layout the subplot grid.

    fig_kw : Other keyword arguments to be passed to the figure() call.
        Note that all keywords not recognized above will be
        automatically included here.

    Returns:

    fig, ax : tuple
      - fig is the Matplotlib Figure object
      - ax can be either a single axis object or an array of axis objects if
      more than one subplot was created.  The dimensions of the resulting array
      can be controlled with the squeeze keyword, see above.

    **Examples:**

    x = np.linspace(0, 2*np.pi, 400)
    y = np.sin(x**2)

    # Just a figure and one subplot
    f, ax = plt.subplots()
    ax.plot(x, y)
    ax.set_title('Simple plot')

    # Two subplots, unpack the output array immediately
    f, (ax1, ax2) = plt.subplots(1, 2, sharey=True)
    ax1.plot(x, y)
    ax1.set_title('Sharing Y axis')
    ax2.scatter(x, y)

    # Four polar axes
    plt.subplots(2, 2, subplot_kw=dict(polar=True))
    """
    import matplotlib.pyplot as plt

    nrows, ncols = _get_layout(naxes, layout=layout, layout_type=layout_type)
    nplots = nrows * ncols

    # Create empty object array to hold all axes.  It's easiest to make it 1-d
    # so we can just append subplots upon creation, and then
    axarr = np.empty(nplots, dtype=object)

    # Create first subplot separately, so we can share it if requested
    ax0 = fig.add_subplot(nrows, ncols, 1, **subplot_kw)

    if sharex:
        subplot_kw["sharex"] = ax0
    if sharey:
        subplot_kw["sharey"] = ax0
    axarr[0] = ax0

    # Note off-by-one counting because add_subplot uses the MATLAB 1-based
    # convention.
    for i in range(1, nplots):
        kwds = subplot_kw.copy()
        # Set sharex and sharey to None for blank/dummy axes, these can
        # interfere with proper axis limits on the visible axes if
        # they share axes e.g. issue #7528
        if i >= naxes:
            kwds["sharex"] = None
            kwds["sharey"] = None
        ax = fig.add_subplot(nrows, ncols, i + 1, **kwds)
        axarr[i] = ax

    if naxes != nplots:
        for ax in axarr[naxes:]:
            ax.set_visible(False)

    _handle_shared_axes(axarr, nplots, naxes, nrows, ncols, sharex, sharey)

    if squeeze:
        # Reshape the array to have the final desired dimension (nrow,ncol),
        # though discarding unneeded dimensions that equal 1.  If we only have
        # one subplot, just return it instead of a 1-element array.
        if nplots == 1:
            axes = axarr[0]
        else:
            axes = axarr.reshape(nrows, ncols).squeeze()
    else:
        # returned axis array will be always 2-d, even if nrows=ncols=1
        axes = axarr.reshape(nrows, ncols)

    return fig, axes


import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np

import gempy_legacy as gp


# ? This test does not seem to be testing what it says it is testing
def test_sort_surfaces_by_solution(one_fault_model_topo_solution):
    geo_model = one_fault_model_topo_solution
    section_dict = {'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                    'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
    geo_model.set_section_grid(section_dict)

    geo_model.set_active_grid('sections', reset=True)

    s1 = geo_model.solutions.scalar_field_at_surface_points
    geo_model.update_additional_data()
    geo_model.update_to_interpolator()
    
    gp.compute_model(geo_model, sort_surfaces=True)
    gp.plot_2d(geo_model, section_names=['section_NW-SE'], show_topography=False)
    plt.show()
    s2 = geo_model.solutions.scalar_field_at_surface_points

    gp.compute_model(geo_model, sort_surfaces=True)
    gp.plot_2d(geo_model, section_names=['section_NW-SE'], show_topography=True)
    plt.show()
    s3 = geo_model.solutions.scalar_field_at_surface_points
    np.testing.assert_array_equal(s2, s3)

    gp.compute_model(geo_model, sort_surfaces=True)
    gp.plot_2d(geo_model, section_names=['section_SW-NE'], show_topography=True)
    plt.show()

    return geo_model



import numpy as np
from .create_topography import LoadDEMArtificial, LoadDEMGDAL
import skimage


class Topography:
    """
    Object to include topography in the model.

    Notes:
        This always assumes that the topography we pass fits perfectly the extent

    """
    def __init__(self, regular_grid=None, regular_grid_extent=None, regular_grid_resolution=None):

        if regular_grid is None and (regular_grid_extent is None or regular_grid_resolution is None):
            raise AttributeError('You need to pass either a regular grid or'
                                 'extent and resolution')

        # Set the extent and resolution of the grid
        if regular_grid_resolution is None:
            self.regular_grid_resolution = regular_grid.resolution[:2]
        else:
            self.regular_grid_resolution = regular_grid_resolution
        assert all(np.asarray(self.regular_grid_resolution) >= 2), 'The regular grid needs to be at least of size 2 on all ' \
                                                      'directions.'
        self.extent = regular_grid.extent[:] if regular_grid_extent is None else regular_grid_extent

        # Values (n, 3)
        self.values = np.zeros((0, 3))

        # Values (n, n, 3)
        self.values_2d = np.zeros((0, 3))

        # Shape original
        self.raster_shape = tuple()

        # Topography Resolution
        self.resolution = np.zeros((0, 3))

        # Source for the
        self.source = None

        # Coords
        self._x = None
        self._y = None

    @property
    def x(self):
        if self._x is not None:
            return self._x
        else:
            val = self.values[:, 0]
            val.sort()
            return np.unique(val)

    @property
    def y(self):
        if self._y is not None:
            return self._y
        else:
            val = self.values[:, 1]
            val.sort()
            return np.unique(val)

    def set_values(self, values_2d: np.ndarray):
        """General method to set topography

        Args:
            values_2d (numpy.ndarray[float,float, 3]): array with the XYZ values
             in 2D

        Returns:
            :class:`gempy.core.grid_modules.topography.Topography`


        """
        # Original topography data
        self.values_2d = values_2d
        self.resolution = values_2d.shape[:2]

        # n,3 array
        self.values = values_2d.reshape((-1, 3), order='C')
        return self

    def crop_topography(self, extent):
        """Crop the topography to a given extent.

        This may be useful for example to mask the regular grid.

        Args:
            extent:

        Returns:

        """
        raise NotImplementedError

    def resize_topo(self):
        regular_grid_topo = skimage.transform.resize(
            self.values_2d,
            (self.regular_grid_resolution[0], self.regular_grid_resolution[1]),
            mode='constant',
            anti_aliasing=False, preserve_range=True)

        return regular_grid_topo

    def load_random_hills(self, **kwargs):
        if 'extent' in kwargs:
            self.extent = kwargs.pop('extent')

        if 'resolution' in kwargs:
            self.regular_grid_resolution = kwargs.pop('resolution')

        dem = LoadDEMArtificial(extent=self.extent,
                                resolution=self.regular_grid_resolution, **kwargs)

        self._x, self._y = dem.x, dem.y
        self.set_values(dem.get_values())

    def load_from_gdal(self, filepath):
        dem = LoadDEMGDAL(filepath, extent=self.extent)
        self._x, self._y = None, None
        self.set_values(dem.get_values())
        # self.source = 'gdal'

    def save(self, path):
        np.save(path, self.values_2d)

    def load(self, path):
        self.set_values(np.load(path))
        self._x, self._y = None, None
        return self.values

    def load_from_saved(self, *args, **kwargs):
        self.load(*args, **kwargs)





import aesara
import aesara.tensor  as tt
import gempy_legacy as gp
import copy


class GemPyThOp:
    def __init__(self, model: gp.Project):
        model = copy.deepcopy(model)
        gp.set_interpolator(model, compile_aesara=False,
                            output=['geology', 'gravity', 'magnetics'],
                            gradient=True)
        self.model = model
        self.th_op = None

    def get_output(self, output):
        if output == 'gravity':
            out = self.model._interpolator.aesara_graph.aesara_output()[12][0]

        elif output == 'lith':
            out = self.model._interpolator.aesara_graph.aesara_output()[0][1]
        else:
            raise AttributeError()

        return out

    def get_wrt(self, wrt: str):
        if wrt == 'surface_points':
            wrt_ = self.model._interpolator.aesara_graph.input_parameters_loop[4]
        else:
            raise AttributeError

        return wrt_

    def set_th_op(self, output):
        interpolator = self.model._interpolator
        out = self.get_output(output)

        i = interpolator.get_python_input_block()
        aesara.config.compute_test_value = 'ignore'
        self.th_op = aesara.OpFromGraph(interpolator.aesara_graph.input_parameters_loop,
                                        [out],
                                        inline=False,
                                        on_unused_input='ignore',
                                        name=output)
        return self.th_op

    def test_gradient(self, output: str, wrt: str):
        aesara.config.compute_test_value = 'ignore'
        interpolator = self.model._interpolator
        out = self.get_output(output)
        wrt_ = self.get_wrt(wrt)

        geo_model_T = aesara.OpFromGraph(interpolator.aesara_graph.input_parameters_loop,
                                         [aesara.grad(out[0], wrt_)],
                                         inline=True,
                                         on_unused_input='ignore',
                                         name='test_'+output)

        i = interpolator.get_python_input_block()
        th_f = aesara.function([], geo_model_T(*i), on_unused_input='warn')

        interpolator.aesara_graph.sig_slope.set_value(20)

        return th_f()

    @staticmethod
    def set_shared(python_input):
        input_sh = []
        i = python_input
        for ii in i:
            input_sh.append(aesara.shared(ii))

        return input_sh


def gempy_th_op(geo_model):

    aesara.config.compute_test_value = 'ignore'
    geo_model_T = aesara.OpFromGraph(geo_model.interpolator.aesara_graph.input_parameters_loop,
                                    [aesara.grad(geo_model.interpolator.aesara_graph.aesara_output()[12],
                                                 geo_model.interpolator.aesara_graph.input_parameters_loop[4])],
                                     inline=True,
                                     on_unused_input='ignore',
                                     name='forw_grav')

    # %%
    i = geo_model.interpolator.get_python_input_block()
    th_f = aesara.function([], geo_model_T(*i), on_unused_input='warn')

    # %%
    geo_model.interpolator.aesara_graph.sig_slope.set_value(20)

    # %%
    th_f()


    # %%
    # Setup Bayesian model
    # --------------------
    #

    # %%
    i = geo_model.interpolator.get_python_input_block()
    aesara.config.compute_test_value = 'ignore'
    geo_model_T_grav = aesara.OpFromGraph(geo_model.interpolator.aesara_graph.input_parameters_loop,
                                    [geo_model.interpolator.aesara_graph.aesara_output()[12]],
                                     inline=False,
                                     on_unused_input='ignore',
                                     name='forw_grav')

    # %%
    geo_model_T_thick = aesara.OpFromGraph(geo_model.interpolator.aesara_graph.input_parameters_loop,
                                    [geo_model.interpolator.aesara_graph.compute_series()[0][1][0:250000]], inline=True,
                                     on_unused_input='ignore',
                                     name='geo_model')

    # %%
    # We convert a python variable to aesara.shared
    input_sh = []
    i = geo_model.interpolator.get_python_input_block()
    for ii in i:
        input_sh.append(aesara.shared(ii))

    # We get the rescaling parameters:
    rf = geo_model.rescaling.df.loc['values', 'rescaling factor'].astype('float32')
    centers = geo_model.rescaling.df.loc['values', 'centers'].astype('float32')

    # We create pandas groups by id to be able to modify several points at the same time:
    g = geo_model.surface_points.df.groupby('id')
    l = aesara.shared(np.array([], dtype='float64'))

    # %%
    g_obs_p = 1e3 * np.array([-0.3548658 , -0.35558686, -0.3563156 , -0.35558686, -0.3548658 ,
           -0.3534237 , -0.35201198, -0.3534237 , -0.3548658 , -0.3563401 ,
           -0.3548658 , -0.35558686, -0.3548658 , -0.3541554 , -0.3534569 ,
           -0.3527707 , -0.35424498, -0.35575098, -0.3572901 , -0.35575098,
           -0.35424498, -0.35575098, -0.35424498, -0.35575098, -0.35424498,
           -0.35575098, -0.35643718, -0.35713565, -0.35643718], dtype='float32')

    y_obs_list = 1e3 * np.array([2.12, 2.06, 2.08, 2.05, 2.08, 2.09,
                  2.19, 2.07, 2.16, 2.11, 2.13, 1.92])

    # %%
    # Python input variables
    i = geo_model.interpolator.get_python_input_block()



from os import path
import sys

# This is for sphenix to find the packages
sys.path.append(path.dirname( path.dirname( path.abspath(__file__) ) ) )

from gempy_legacy.core.model import *
from typing import Union


def add_surface_points(data: Union[Project, SurfacePoints]):
    pass


def add_orienations(data: Union[Project, Orientations]):
    pass


def add_series(data: Union[Project, Series]):
    pass


def add_formations(data: Union[Project, Surfaces]):
    pass

def del_surface_points(data: Union[Project, SurfacePoints]):
    pass


def del_orientations(data: Union[Project, Orientations]):
    pass


def del_series(data: Union[Project, Series]):
    pass


def del_formations(data: Union[Project, Surfaces]):
    pass


def modify_surface_points(data: Union[Project, SurfacePoints]):
    pass

def modify_orientations(data: Union[Project, Orientations]):
    pass


def modify_series(data: Union[Project, Series]):
    pass

def modify_formations(data: Union[Project, Surfaces]):
    pass

def modify_faults(data: Union[Project, Faults]):
    pass

def set_is_fault(data: Union[Project, Faults], idx):
    if isinstance(data, Project):
        data.set_is_fault(idx)
    elif isinstance(data, Faults):
        data.set_is_fault()

def modify_faults_network(data: Union[Project, Faults]):
    pass

def modify_options(data: Union[Project, Options]):
    pass

def modify_kriging_parameters(data: Union[Project, KrigingParameters]):
    pass

def modify_rescaling_parametesr(data: Union[Project, ScalingSystem]):
    pass



"""
Chapter 4: Analyzing Geomodel Topology
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import gempy_legacy as gp
from gempy_legacy.assets import topology as tp

import numpy as np
import matplotlib.pyplot as plt

import warnings
warnings.filterwarnings("ignore")


# %%
# Load example Model
# ^^^^^^^^^^^^^^^^^^
# 
# First let's set up a very simple example model. For that we initialize
# the geo_data object with the correct model extent and the resolution we
# like. Then we load our data points from csv files and set the series and
# order the formations (stratigraphic pile).
# 

# %% 
geo_model = gp.create_model("Model_Tutorial6")

data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
gp.init_data(
    geo_model, [0, 3000, 0, 20, 0, 2000], [50, 10, 67], 
    path_i=data_path+"data/input_data/tut_chapter6/ch6_data_interf.csv", 
    path_o=data_path+"data/input_data/tut_chapter6/ch6_data_fol.csv"
)
gp.map_stack_to_surfaces(
    geo_model,
    {
        "fault": "Fault",
        "Rest": ('Layer 2', 'Layer 3', 'Layer 4', 'Layer 5')
    }
)
geo_model.set_is_fault(["fault"]);
gp.set_interpolator(geo_model)
sol = gp.compute_model(geo_model, compute_mesh=True)

# %% 
gp.plot_2d(geo_model, cell_number=[5])


# %%
# Analyzing Topology
# ^^^^^^^^^^^^^^^^^^
# 
# GemPy sports in-built functionality to analyze the topology of its
# models. All we need for this is our geo_data object, lithology block and
# the fault block. We input those into *gp.topology_compute* and get
# several useful outputs:
# 
# -  an adjacency graph **G**, representing the topological relationships
#    of the model
# -  the **centroids** of the all the unique topological regions in the
#    model (x,y,z coordinates of their center)
# -  a list of all the unique labels (labels_unique)
# -  two look-up-tables from the lithology id's to the node labels, and
#    vice versa
# 

# %% 
edges, centroids = tp.compute_topology(geo_model)


# %%
# The first output of the topology function is the ``set`` of edges
# representing topology relationships between unique geobodies of the
# block model. An edge is represented by a ``tuple`` of two ``int``
# geobody (or node) labels:
# 

# %% 
edges


# %%
# The second output is the centroids ``dict``, mapping the unique geobody
# id's (graph node id's) to the geobody centroid position in grid
# coordinates:
# 

# %% 
centroids


# %%
# After computing the model topology, we can overlay the topology graph
# over a model section:
# 


# %%
# Visualizing topology
# ~~~~~~~~~~~~~~~~~~~~
# 
# 2-D Visualization of the Topology Graph
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 

# %% 
gp.plot.plot_topology(geo_model, edges, centroids)
plt.show()

# %% 
gp.plot_2d(geo_model, cell_number=[5], show=False)
gp.plot.plot_topology(geo_model, edges, centroids, scale=True)
plt.show()

# %%
# Adjacency Matrix
# ~~~~~~~~~~~~~~~~
# 
# Another way to encode and visualize the geomodel topology is using an
# adjacency graph:
# 

# %% 
M = tp.get_adjacency_matrix(geo_model, edges, centroids)
print(M)

# %% 
tp.plot_adjacency_matrix(geo_model, M)


# %%
# 3-D Visualization of the Topology Graph
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 


# %%
# You can also plot the topology in 3-D using GemPy's 3-D visualization
# toolkit powered by ``pyvista``:
# 

# %% 
from gempy_legacy.plot._vista import Vista
gpv = Vista(geo_model)
gpv.plot_topology(edges, centroids)
gpv.show()


# %%
# Look-up tables
# ~~~~~~~~~~~~~~
# 


# %%
# The ``topology`` asset provides several look-up tables to work with the
# unique geobody topology id's.
# 
# Mapping node id's back to lithology / surface id's:
# 

# %% 
lith_lot = tp.get_lot_node_to_lith_id(geo_model, centroids)
lith_lot


# %%
# Figuring out which nodes are in which fault block:
# 

# %% 
fault_lot = tp.get_lot_node_to_fault_block(geo_model, centroids)
fault_lot


# %%
# We can also easily map the lithology id to the corresponding topology
# id's:
# 

# %% 
tp.get_lot_lith_to_node_id(lith_lot)


# %%
# Detailed node labeling
# ~~~~~~~~~~~~~~~~~~~~~~
# 

# %%
# sphinx_gallery_thumbnail_number = 4
dedges, dcentroids = tp.get_detailed_labels(geo_model, edges, centroids)
gp.plot_2d(geo_model, cell_number=[5], show=False)
gp.plot.plot_topology(geo_model, dedges, dcentroids, scale=True)
plt.show()

# %% 
dedges

# %% 
dcentroids


# %%
# Checking adjacency
# ~~~~~~~~~~~~~~~~~~
# 


# %%
# So lets say we want to check if the purple layer (id 5) is connected
# across the fault to the yellow layer (id 3). For this we can make easy
# use of the detailed labeling and the ``check_adjacency`` function:
# 

# %% 
tp.check_adjacency(dedges, "5_1", "3_0")


# %%
# We can also check all geobodies that are adjacent to the purple layer
# (id 5) on the left side of the fault (fault id 1):
# 

# %% 
tp.get_adjacencies(dedges, "5_1")

# %% 
